@@ -1,5 +1,5 @@
 -- hkt.hs: hOpenPGP key tool
--- Copyright © 2013-2015  Clint Adams
+-- Copyright © 2013-2016  Clint Adams
 --
 -- vim: softtabstop=4:shiftwidth=4:expandtab
 --
@@ -162,7 +162,7 @@ listO homedir = Options
        <> showDefault
        <> help "output format" )
     <*> switch ( long "filter" <> help "treat target as filter" )
-    <*> argument str ( metavar "TARGET" <> targetHelp )
+    <*> (fromMaybe "" <$> optional (argument str ( metavar "TARGET" <> targetHelp )))
     <*> pure ""
     <*> pure ""
     where
@@ -182,7 +182,7 @@ graphO homedir = Options
        <> ofhelp )
     <*> pure Unstructured -- unused
     <*> switch ( long "filter" <> help "treat target as filter" )
-    <*> argument str ( metavar "TARGET" <> targetHelp )
+    <*> (fromMaybe "" <$> optional (argument str ( metavar "TARGET" <> targetHelp )))
     <*> pure ""
     <*> pure ""
     where
