@@ -1,5 +1,5 @@
 -- hokey.hs: hOpenPGP key tool
--- Copyright © 2013-2015  Clint Adams
+-- Copyright © 2013-2016  Clint Adams
 --
 -- vim: softtabstop=4:shiftwidth=4:expandtab
 --
@@ -19,9 +19,13 @@
 {-# LANGUAGE DeriveFunctor, DeriveGeneric, FlexibleInstances #-}
 
 import HOpenPGP.Tools.Common (banner, versioner, warranty)
+import HOpenPGP.Tools.HKP (fetchKeys, FetchValidationMethod(..), rearmorKeys)
+import HOpenPGP.Tools.TKUtils (processTK)
+import qualified Codec.Encryption.OpenPGP.ASCIIArmor as AA
 import Codec.Encryption.OpenPGP.Expirations (getKeyExpirationTimesFromSignature)
 import Codec.Encryption.OpenPGP.Fingerprint (fingerprint, eightOctetKeyID)
 import Codec.Encryption.OpenPGP.KeyInfo (pubkeySize, pkalgoAbbrev)
+import Codec.Encryption.OpenPGP.KeySelection (parseFingerprint)
 import Codec.Encryption.OpenPGP.Serialize ()
 import Codec.Encryption.OpenPGP.Signatures (verifyTKWith, verifySigWith, verifyAgainstKeys)
 import Codec.Encryption.OpenPGP.Types
@@ -29,6 +33,7 @@ import Control.Applicative ((<$>),(<*>), pure)
 import Control.Arrow ((***), second)
 import Control.Error.Util (hush)
 import Control.Lens ((^.), (&), _1, _2, mapped, over)
+import Control.Monad.Trans.Except (runExceptT)
 import Control.Monad.Trans.Resource (runResourceT)
 import Control.Monad.Trans.Writer.Lazy (execWriter, tell)
 import qualified Crypto.Hash.SHA3 as SHA3
@@ -57,16 +62,16 @@ import Data.Time.Format (formatTime)
 import qualified Data.Yaml as Y
 import GHC.Generics
 
-import Options.Applicative.Builder (auto, command, footerDoc, headerDoc, helpDoc, info, long, metavar, prefs, progDesc, showDefault, showHelpOnError, option, subparser, value)
+import Options.Applicative.Builder (argument, auto, command, eitherReader, footerDoc, headerDoc, help, helpDoc, info, long, metavar, prefs, progDesc, showDefault, showHelpOnError, option, subparser, str, value)
 import Options.Applicative.Extra (customExecParser, helper)
 import Options.Applicative.Types (Parser)
 
-import System.IO (Handle, hFlush, stderr, stdin, stdout, hSetBuffering, BufferMode(..))
+import System.IO (Handle, hFlush, hPutStrLn, stderr, stdin, stdout, hSetBuffering, BufferMode(..))
 import Data.Time.Locale.Compat (defaultTimeLocale)
 
 import qualified Text.PrettyPrint.ANSI.Leijen as PPAL
 import Text.PrettyPrint.Free (colon, indent, linebreak, list, pretty, text, (<+>), Doc)
-import System.Console.Terminfo.PrettyPrint (green, red, yellow, displayLn, Effect)
+import System.Console.Terminfo.PrettyPrint (green, red, yellow, displayDoc, Effect)
 
 data KAS = KAS {
     pubkeyalgo :: Colored PubKeyAlgorithm
@@ -93,6 +98,8 @@ data KeyReport = KeyReport {
   , keyAlgorithmAndSize :: KAS
   , keyUIDsAndUAts :: FakeMap Text (Colored UIDReport)
   , keyBestOf :: Maybe UIDReport
+  , keySubkeys :: [SubkeyReport]
+  , keyHasEncryptionCapableSubkey :: Colored Bool
 } deriving Generic
 
 data UIDReport = UIDReport {
@@ -103,6 +110,21 @@ data UIDReport = UIDReport {
   , uidRevocationStatus :: [RevocationStatus]
 } deriving Generic
 
+data SubkeyReport = SubkeyReport {
+    skFingerprint :: TwentyOctetFingerprint
+  , skVer :: Colored KeyVersion
+  , skCreationTime :: ThirtyTwoBitTimeStamp
+  , skAlgorithmAndSize :: KAS
+  , skBindingSigHashAlgorithms :: [Colored HashAlgorithm]
+  , skUsageFlags :: [Colored (Set.Set KeyFlag)]
+  , skCrossCerts :: CrossCertReport
+} deriving Generic
+
+data CrossCertReport = CrossCertReport {
+    ccPresent :: Colored Bool
+  , ccHashAlgorithms :: [Colored HashAlgorithm]
+} deriving Generic
+
 data RevocationStatus = RevocationStatus {
     isRevoked :: Bool
   , revocationCode :: String
@@ -114,6 +136,8 @@ instance A.ToJSON Color
 instance (A.ToJSON a) => A.ToJSON (Colored a)
 instance A.ToJSON KeyReport
 instance A.ToJSON UIDReport
+instance A.ToJSON SubkeyReport
+instance A.ToJSON CrossCertReport
 instance A.ToJSON b => A.ToJSON (FakeMap Text b) where
     toJSON = A.toJSON . Map.fromList . unFakeMap
 instance A.ToJSON RevocationStatus
@@ -123,22 +147,26 @@ instance Monoid UIDReport where
     mappend (UIDReport a b c d e) (UIDReport a' b' c' d' e') = UIDReport (a <> a') (b <> b') (c <> c') (d <> d') (e <> e')
 
 checkKey :: Maybe POSIXTime -> TK -> KeyReport
-checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x }) KeyReport {
+checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x, keyHasEncryptionCapableSubkey = hasEncryptionCapableSubkey (concatMap skUsageFlags (keySubkeys x)) }) KeyReport {
     keyStatus = either id (const "good") processedTK
   , keyFingerprint = key^.tkKey._1 & fingerprint
   , keyVer = key^.tkKey._1.keyVersion & colorizeKV
   , keyCreationTime = key^.tkKey._1.timestamp
-  , keyAlgorithmAndSize = KAS {
-      pubkeyalgo = key^.tkKey._1.pkalgo & colorizePKA
-    , pubkeysize = key^.tkKey._1.pubkey & colorizePKS . pubkeySize
-    , stringrep = (key^.tkKey._1.pubkey & either (const "unknown") show . pubkeySize) ++ (key^.tkKey._1.pkalgo & pkalgoAbbrev)
-  }
+  , keyAlgorithmAndSize = kasIt (key^.tkKey._1)
   , keyUIDsAndUAts = FakeMap (map (\(x,y) -> (x, uidr (Just x) y)) (processedOrOrig^.tkUIDs) ++ map (uatspsToText *** uidr Nothing) (processedOrOrig^.tkUAts))
   , keyBestOf = Nothing
+  , keySubkeys = map checkSK (key^.tkSubs)
+  , keyHasEncryptionCapableSubkey = Colored Nothing Nothing False
 }
     where
         processedOrOrig = either (const key) id processedTK
-        processedTK = verifyTKWith (verifySigWith (verifyAgainstKeys [key])) (fmap posixSecondsToUTCTime mpt) . stripOlderSigs . stripOtherSigs $ key
+        processedTK = processTK mpt key
+        kasIt :: PKPayload -> KAS
+        kasIt pkp = KAS {
+          pubkeyalgo = pkp^.pkalgo & colorizePKA
+        , pubkeysize = pkp^.pubkey & colorizePKS . pubkeySize
+        , stringrep = (pkp^.pubkey & either (const "unknown") show . pubkeySize) ++ (pkp^.pkalgo & pkalgoAbbrev)
+        }
         colorizeKV kv = uncurry Colored (if kv == V4 then (Just Green, Nothing) else (Just Red, Just "not a V4 key")) kv
         colorizePKA pka = uncurry Colored (if pka == RSA then (Just Green, Nothing) else (Just Yellow, Just "not an RSA key")) pka
         colorizePKS pks = uncurry Colored (colorizePKS' pks) (hush pks)
@@ -170,15 +198,6 @@ checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x }) KeyReport {
         isCT _ = False
         sigcts (SigV4 _ _ _ xs _ _ _) = map (\(SigSubPacket _ (SigCreationTime x)) -> x) $ filter isCT xs
         alleged = filter (\x -> ((==) <$> sigissuer x <*> eoki (key^.tkKey._1)) == Just True)
-        newest = take 1 . sortBy (comparing (Down . take 1 . sigcts)) -- FIXME: this is terrible
-        stripOtherSigs tk = tk {
-            _tkUIDs = map (second alleged) (_tkUIDs tk)
-          , _tkUAts = map (second alleged) (_tkUAts tk)
-        }
-        stripOlderSigs tk = tk {
-            _tkUIDs = map (second newest) (_tkUIDs tk)
-          , _tkUAts = map (second newest) (_tkUAts tk)
-        }
         uatspsToText = T.pack . uatspsToString
         uatspsToString us = "<uat:[" ++ intercalate "," (map uaspToString us) ++ "]>"
         uaspToString (ImageAttribute hdr d) = hdrToString hdr ++ ':':show (BL.length d) ++ ':':BC8.unpack (Base16.encode (SHA3.hashlazy 48 d))
@@ -188,12 +207,12 @@ checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x }) KeyReport {
         uidr Nothing sps = Colored Nothing Nothing (UIDReport (has sps)
                                                               (map phas sps)
                                                               (map (colorizeKETs (fromMaybe 0 mpt) (key^.tkKey._1.timestamp & unThirtyTwoBitTimeStamp) . getKeyExpirationTimesFromSignature) sps) -- should that be 0?
-                                                              (kufs sps)
+                                                              (kufs False sps)
                                                               (findRevocationReason sps))
         uidr (Just u) sps = colorizeUID u (UIDReport (has sps)
                                                      (map phas sps)
                                                      (map (colorizeKETs (fromMaybe 0 mpt) (key^.tkKey._1.timestamp & unThirtyTwoBitTimeStamp) . getKeyExpirationTimesFromSignature) sps) -- should that be 0?
-                                                     (kufs sps)
+                                                     (kufs False sps)
                                                      (findRevocationReason sps))
         populateBestOf krep = Just (UIDReport <$> best . uidSelfSigHashAlgorithms <*> best . uidPreferredHashAlgorithms <*> best . uidKeyExpirationTimes <*> best . uidKeyUsageFlags <*> pure [] $ mconcat (justTheUIDRs krep))
         justTheUIDRs = map (decolorize . snd) . unFakeMap . keyUIDsAndUAts
@@ -210,24 +229,60 @@ checkKey mpt key = (\x -> x { keyBestOf = populateBestOf x }) KeyReport {
         grabReasons _ = []
         grabReasons' (ReasonForRevocation a b) = Just (RevocationStatus True (show a) b)
         grabReasons' _ = Nothing
-        kufs = map (colorizeKUFs . (\(SigSubPacket _ (KeyFlags x)) -> x) . maybe undefined id . find isKUF . hasheds) . newestWith (any isKUF . hasheds)
+        kufs s = map (colorizeKUFs s . (\(SigSubPacket _ (KeyFlags x)) -> x) . maybe undefined id . find isKUF . hasheds) . newestWith (any isKUF . hasheds)
         isKUF (SigSubPacket _ (KeyFlags _)) = True
         isKUF _ = False
-        colorizeKUFs x = uncurry Colored (if (Set.member EncryptStorageKey x || Set.member EncryptCommunicationsKey x) && (Set.member SignDataKey x || Set.member CertifyKeysKey x) then (Just Yellow, Just "both signing & encryption") else (Just Green, Nothing)) x
+        colorizeKUFs False x = uncurry Colored (if (Set.member EncryptStorageKey x || Set.member EncryptCommunicationsKey x) && (Set.member SignDataKey x || Set.member CertifyKeysKey x) then (Just Yellow, Just "both signing & encryption") else (Just Green, Nothing)) x
+        colorizeKUFs True x = uncurry Colored (if (Set.member CertifyKeysKey x) then (Just Red, Just "certification-capable subkey") else (if (Set.member EncryptStorageKey x || Set.member EncryptCommunicationsKey x) && Set.member SignDataKey x then (Just Yellow, Just "both signing & encryption") else (Just Green, Nothing))) x
         newestWith p = take 1 . sortBy (comparing (Down . take 1 . sigcts)) . filter p -- FIXME: this is terrible
         hasheds (SigV4 _ _ _ xs _ _ _) = xs
         hasheds _ = []
+        checkSK :: (Pkt, [SignaturePayload]) -> SubkeyReport
+        checkSK (PublicSubkeyPkt pkp, sigs) = checkSK' pkp sigs
+        checkSK (SecretSubkeyPkt pkp _, sigs) = checkSK' pkp sigs
+        checkSK' pkp sigs = (\x -> x { skCrossCerts = ccr (map decolorize (skUsageFlags x)) sigs }) SubkeyReport {
+          skFingerprint = fingerprint pkp
+        , skVer = colorizeKV (pkp^.keyVersion)
+        , skCreationTime = pkp^.timestamp
+        , skAlgorithmAndSize = kasIt pkp
+        , skBindingSigHashAlgorithms = has (filter isSKBindingSig sigs)
+        , skUsageFlags = kufs True (filter isSKBindingSig sigs)
+        , skCrossCerts = CrossCertReport (Colored Nothing Nothing False) []
+        }
+        hasEncryptionCapableSubkey skrs = if any ((\x -> Set.member EncryptStorageKey x || Set.member EncryptCommunicationsKey x) . decolorize) skrs then Colored (Just Green) Nothing True else Colored (Just Red) (Just "no encryption-capable subkey present") False
+        isSKBindingSig (SigV4 SubkeyBindingSig _ _ _ _ _ _) = True
+        isSKBindingSig _ = False
+        isPKBindingSig (SigV4 PrimaryKeyBindingSig _ _ _ _ _ _) = True
+        isPKBindingSig _ = False
+        embeddedSigs = filter isPKBindingSig . concatMap getEmbeds . filter isSKBindingSig
+        getEmbeds (SigV4 _ _ _ _ xs _ _) = concatMap getEmbed xs
+        getEmbeds _ = []
+        getEmbed (SigSubPacket _ (EmbeddedSignature sp)) = [sp]
+        getEmbed _ = []
+        ccr kufs sigs = CrossCertReport (colorES kufs sigs) (map (colorizeHA . hashAlgo) sigs)
+        colorES kufs sigs = case (null (embeddedSigs sigs), any (Set.member SignDataKey) kufs, any (Set.member AuthKey) kufs) of
+                                (True, True, True) -> Colored (Just Red) (Just "signing- and auth-capable subkey without cross-cert") False
+                                (True, True, False) -> Colored (Just Red) (Just "signing-capable subkey without cross-cert") False
+                                (True, False, True) -> Colored (Just Yellow) (Just "auth-capable subkey without cross-cert") False
+                                (False, True, True) -> Colored (Just Green) Nothing True
+                                (False, True, False) -> Colored (Just Green) Nothing True
+                                (False, False, True) -> Colored (Just Green) Nothing True
+                                (False, False, False) -> Colored Nothing Nothing True
+                                (True, _, _) -> Colored Nothing Nothing False
 
 prettyKeyReport :: POSIXTime -> TK -> Doc Effect
 prettyKeyReport cpt key = do
  let keyReport = checkKey (Just cpt) key
  execWriter $ tell
-   ( linebreak <> text "Key has potential validity:" <+> text (keyStatus keyReport)
-  <> linebreak <> text "Key has fingerprint:" <+> pretty (SpacedFingerprint (keyFingerprint keyReport))
-  <> linebreak <> text "Checking to see if key is OpenPGPv4:" <+> coloredToColor (text . show) (keyVer keyReport)
-  <> linebreak <> (\kas -> text "Checking to see if key is RSA or DSA (>= 2048-bit):" <+> coloredToColor pretty (pubkeyalgo kas) <+> coloredToColor (maybe (text "unknown") pretty) (pubkeysize kas)) (keyAlgorithmAndSize keyReport)
-  <> linebreak <> text "Checking user-ID- and user-attribute-related items:"
+   ( linebreak <> text "Key has potential validity" <> colon <+> text (keyStatus keyReport)
+  <> linebreak <> text "Key has fingerprint" <> colon <+> pretty (SpacedFingerprint (keyFingerprint keyReport))
+  <> linebreak <> text "Checking to see if key is OpenPGPv4" <> colon <+> coloredToColor (text . show) (keyVer keyReport)
+  <> linebreak <> (\kas -> text "Checking to see if key is RSA or DSA (>= 2048-bit)" <> colon <+> coloredToColor pretty (pubkeyalgo kas) <+> coloredToColor (maybe (text "unknown") pretty) (pubkeysize kas)) (keyAlgorithmAndSize keyReport)
+  <> linebreak <> text "Checking user-ID- and user-attribute-related items" <> colon
   <> mconcat (map (uidtrip (keyCreationTime keyReport) . gottabeabetterway) (unFakeMap (keyUIDsAndUAts keyReport)))
+  <> linebreak <> text "Checking subkeys" <> colon
+  <> linebreak <> indent 2 (text "one of the subkeys is encryption-capable" <> colon <+> coloredToColor pretty (keyHasEncryptionCapableSubkey keyReport))
+  <> mconcat (map (subkeyrep) (keySubkeys keyReport))
   <> linebreak
    )
     where
@@ -246,6 +301,14 @@ prettyKeyReport cpt key = do
           <> linebreak <> indent 4 (text "Revocation reason" <> colon <+> list (map (text . T.unpack . revocationReason) (uidRevocationStatus ur)))
         keyExp ts ke = (show . pretty) ke ++ " = " ++ formatTime defaultTimeLocale "%c" (posixSecondsToUTCTime (realToFrac ts + realToFrac ke))
         gottabeabetterway (a, Colored x y z) = (Colored x y a, z)
+        subkeyrep skr = linebreak <> indent 2 (text "fpr" <> colon <+> pretty (SpacedFingerprint (skFingerprint skr)))
+                     <> linebreak <> indent 4 (text "version" <> colon <+> coloredToColor pretty (skVer skr))
+                     <> linebreak <> indent 4 (text "timestamp" <> colon <+> pretty (skCreationTime skr))
+                     <> linebreak <> indent 4 ((\kas -> text "algo/size" <> colon <+> coloredToColor pretty (pubkeyalgo kas) <+> coloredToColor (maybe (text "unknown") pretty) (pubkeysize kas)) (skAlgorithmAndSize skr))
+                     <> linebreak <> indent 4 (text "binding sig hash algorithms" <> colon <+> (list . map (coloredToColor pretty) . skBindingSigHashAlgorithms) skr)
+                     <> linebreak <> indent 4 (text "usage flags" <> colon <+> (list . map (coloredToColor (pretty . Set.toList))) (skUsageFlags skr))
+                     <> linebreak <> indent 4 (text "embedded cross-cert" <> colon <+> (coloredToColor pretty . ccPresent . skCrossCerts) skr)
+                     <> linebreak <> indent 4 (text "cross-cert hash algorithms" <> colon <+> (list . map (coloredToColor pretty) . ccHashAlgorithms . skCrossCerts) skr)
 
 jsonReport :: POSIXTime -> TK -> BL.ByteString
 jsonReport ps = A.encode . checkKey (Just ps)
@@ -260,7 +323,13 @@ data LintOptions = LintOptions {
     lintOutputFormat :: LintOutputFormat
 }
 
-data Command = CmdLint LintOptions | CmdCanonicalize
+data FetchOptions = FetchOptions {
+    keyServer :: String
+  , fetchValidation :: FetchValidationMethod
+  , fetchQuery :: TwentyOctetFingerprint
+}
+
+data Command = CmdLint LintOptions | CmdCanonicalize | CmdFetch FetchOptions
 
 lintO :: Parser LintOptions
 lintO = LintOptions
@@ -275,8 +344,30 @@ lintO = LintOptions
         ofHelp = helpDoc . Just $ PPAL.text "output format" <> PPAL.hardline <> PPAL.list (map (PPAL.text . show) ofchoices)
         ofchoices = [minBound..maxBound] :: [LintOutputFormat]
 
+fetchO :: Parser FetchOptions
+fetchO = FetchOptions
+    <$> option str
+        ( long "keyserver"
+       <> metavar "URL"
+       <> value "http://pool.sks-keyservers.net:11371"
+       <> showDefault
+       <> help "HKP server"
+        )
+    <*> option auto
+        ( long "validation-method"
+       <> metavar "METHOD"
+       <> value MatchPrimaryKeyFingerprint
+       <> showDefault
+       <> vmHelp
+        )
+    <*> argument (eitherReader strToFP) ( metavar "FINGERPRINT" )
+    where
+        vmHelp = helpDoc . Just $ PPAL.text "validation method" <> PPAL.hardline <> PPAL.list (map (PPAL.text . show) vmchoices)
+        vmchoices = [minBound..maxBound] :: [FetchValidationMethod]
+        strToFP = parseFingerprint . T.pack
 
 dispatch :: Command -> IO ()
+dispatch (CmdFetch o) = banner' stderr >> hFlush stderr >> doFetch o
 dispatch (CmdLint o) = banner' stderr >> hFlush stderr >> doLint o
 dispatch (CmdCanonicalize) = banner' stderr >> hFlush stderr >> doCanonicalize
 
@@ -288,6 +379,7 @@ main = do
 cmd :: Parser Command
 cmd = subparser
     ( command "canonicalize" (info ( pure CmdCanonicalize) ( progDesc "arrange key components in a canonical ordering" ))
+   <> command "fetch" (info ( CmdFetch <$> fetchO) ( progDesc "fetch key(s) from keyserver" ))
    <> command "lint" (info ( CmdLint <$> lintO) ( progDesc "check key(s) for 'best practices'" ))
     )
 
@@ -297,7 +389,7 @@ doLint o = do
     keys <- runResourceT $ CB.sourceHandle stdin $= conduitGet get $= conduitToTKsDropping $$ CL.consume
     output (lintOutputFormat o) cpt keys
     where
-        output Pretty cpt = mapM_ (displayLn . prettyKeyReport cpt)
+        output Pretty cpt = mapM_ (displayDoc 0.8 . prettyKeyReport cpt)
         output JSON cpt = mapM_ (BL.putStr . flip BL.append (BL.singleton 0x0a) . jsonReport cpt)
         output YAML cpt = mapM_ (B.putStr . yamlReport cpt)
 
@@ -315,6 +407,13 @@ doCanonicalize = do
         indepthsort :: (Ord a, Ord b) => [(a,[b])] -> [(a,[b])]
         indepthsort = nub . sort . over (mapped._2) sort
 
+doFetch :: FetchOptions -> IO ()
+doFetch o = do
+    ekeys <- runExceptT $ fetchKeys (keyServer o) (fetchValidation o) (fetchQuery o)
+    case ekeys of
+        Left e -> hPutStrLn stderr $ "error fetching keys: " ++ e
+        Right ks -> B.putStr $ rearmorKeys ks
+
 banner' :: Handle -> IO ()
 banner' h = PPAL.hPutDoc h (banner "hokey" <> PPAL.hardline <> warranty "hokey" <> PPAL.hardline)
 
